* { font-size: 16px; line-height: 1.428; text-align: justify; text-justify: inter-word;}

.navbar { min-height:38px; }
.navbar .navbar-brand{padding: 0 20px; font-size: 16px; line-height: 38px; height: 38px; font-weight: bold}
.navbar .navbar-nav > li > a {  font-weight: bold}

.OMFIThead { border-style: solid;
             border-color: #ff0000;
             opacity: 0.75;
             padding-top: 25px;
             padding-right: 50px;
             padding-bottom: 25px;
             padding-left: 50px;
             margin-right: 15px;
             margin-left: 15px;
      -webkit-border-radius: 10px !important;
         -moz-border-radius: 10px !important;
              border-radius: 10px !important;
      -webkit-border-top: 10px !important;
         -moz-border-top: 10px !important;
              border-top: 10px !important;
      -webkit-border-bottom: 10px !important;
         -moz-border-bottom: 10px !important;
              border-bottom: 10px !important;
    }

.OMFITbrd {  margin-right: 15px;
             margin-left: 15px;
}

br {
    border-bottom:1px dashed black;
    display:block;
}