try:
    from test_omas_core import *
    from test_omas_physics import *
    from test_omas_plot import *
    from test_omas_suite import *
    from test_omas_utils import *
except ImportError:
    pass
